// Created by Michael Hoffman. Feel free to use or modify this code as you like. Credit is 
// appreciated but not necessary. Email: hoffman@nullcandy.com

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;
using Microsoft.Xna.Framework.Media;


namespace Metaballs
{
    public class MetaballDemo : Microsoft.Xna.Framework.Game
    {
        public static GraphicsDevice Device;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont font;

        Texture2D instructions, overlay;
        Texture2D[] metaballTextures;
        Color[] glowColors;
        RenderTarget2D metaballTarget;
        AlphaTestEffect effect;

        List<Metaball> balls = new List<Metaball>();
        bool showOverlay = true;
        int currentColor = 0;

        Random rand;
        KeyboardState state, lastState;

#if WINDOWS_PHONE
        public const int MetaballRadius = 32;
        public const float MetaballScale = 4f;
        const int NumMetaballs = 20;
#else
        public const int MetaballRadius = 128;
        public const float MetaballScale = 1f;
        const int NumMetaballs = 50;
#endif

        public MetaballDemo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

#if WINDOWS_PHONE
            IsFixedTimeStep = false;            
            graphics.PreferredBackBufferHeight = 300;
            graphics.PreferredBackBufferWidth = 500;
            graphics.PreferredBackBufferFormat = SurfaceFormat.Color;
            TouchPanel.EnabledGestures = GestureType.Flick;
#endif
        }

        protected override void Initialize()
        {
            base.Initialize();

            // create a bunch of metaballs
            rand = new Random();
            for (int i = 0; i < NumMetaballs; i++)
            {
                var ball = new Metaball();
                ball.Position = new Vector2(rand.Next(GraphicsDevice.Viewport.Width), rand.Next(GraphicsDevice.Viewport.Height)) - new Vector2(128);
                ball.Velocity = rand.NextVector2(1);
                ball.Texture = metaballTextures[0];
                ball.Glow = glowColors[0];
                balls.Add(ball);
            }
        }

        protected override void LoadContent()
        {
            Device = GraphicsDevice;
            spriteBatch = new SpriteBatch(GraphicsDevice);
            font = Content.Load<SpriteFont>("Font");
            instructions = Content.Load<Texture2D>("Instructions");
            overlay = Content.Load<Texture2D>("Overlay");

            // generate some differently coloured metaball textures. You could tint them with SpriteBatch instead, 
            // but then you couldn't have a two colour gradient.
            metaballTextures = new Texture2D[5];
            metaballTextures[0] = Metaball.GenerateTexture(MetaballRadius, Metaball.CreateTwoColorPicker(Color.Red, Color.Yellow));
            metaballTextures[1] = Metaball.GenerateTexture(MetaballRadius, Metaball.CreateTwoColorPicker(Color.Blue, Color.Cyan));
            metaballTextures[2] = Metaball.GenerateTexture(MetaballRadius, Metaball.CreateTwoColorPicker(Color.Green, Color.Yellow));
            metaballTextures[3] = Metaball.GenerateTexture(MetaballRadius, Metaball.CreateTwoColorPicker(Color.Magenta, Color.White));
            metaballTextures[4] = Metaball.GenerateTexture(MetaballRadius, Metaball.CreateTwoColorPicker(Color.Red, Color.Red));

            glowColors = new Color[] { Color.Red, Color.Blue, Color.Lime, Color.Magenta, Color.Red };

            // initialize the alpha test effect.
            effect = new AlphaTestEffect(GraphicsDevice);
            var viewport = GraphicsDevice.Viewport;
            effect.Projection = Matrix.CreateTranslation(-0.5f, -0.5f, 0) * Matrix.CreateOrthographicOffCenter(0, viewport.Width, viewport.Height, 0, 0, 1);
            effect.ReferenceAlpha = 128;

            metaballTarget = new RenderTarget2D(GraphicsDevice, GraphicsDevice.Viewport.Width, GraphicsDevice.Viewport.Height);
        }

        protected override void Update(GameTime gameTime)
        {
            lastState = state;
            state = Keyboard.GetState();

            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed || 
                WasPressed(Keys.Escape))
                this.Exit();

#if WINDOWS
            HandleInputOnPC();
#endif

#if WINDOWS_PHONE
            // On Windows Phone, swiping left or right changes the colors.
            while (TouchPanel.IsGestureAvailable)
            {
                var gesture = TouchPanel.ReadGesture();
                if (gesture.GestureType == GestureType.Flick)
                {
                    int direction = gesture.Delta.X < 0 ? 1 : -1;

                    currentColor = (currentColor + direction) % (metaballTextures.Length + 1);
                    if (currentColor < 0)
                        currentColor += metaballTextures.Length + 1;
#else
            // On PC, pressing 'C' changes the colors.
            if (WasPressed(Keys.C))
            {
                currentColor = (currentColor + 1) % (metaballTextures.Length + 1);
#endif
                    if (currentColor < metaballTextures.Length)
                    {
                        foreach (var ball in balls)
                        {
                            ball.Texture = metaballTextures[currentColor];
                            ball.Glow = glowColors[currentColor];
                        }
                    }
                    else
                    {
                        // The last color option is to mix all the colors.
                        foreach (var ball in balls)
                        {
                            int index = rand.Next(metaballTextures.Length);
                            ball.Texture = metaballTextures[index];
                            ball.Glow = glowColors[index];
                        }
                    }
                }
#if WINDOWS_PHONE
            }
#endif

            foreach (var ball in balls)
                ball.Update();

            base.Update(gameTime);
        }

        private void HandleInputOnPC()
        {
            // allow user to toggle fullscreen mode
            if (Shift && WasPressed(Keys.Enter))
            {
                if (graphics.IsFullScreen)
                {
                    graphics.PreferredBackBufferWidth = 800;
                    graphics.PreferredBackBufferHeight = 480;
                }
                else
                {
                    graphics.PreferredBackBufferWidth = GraphicsDevice.DisplayMode.Width;
                    graphics.PreferredBackBufferHeight = GraphicsDevice.DisplayMode.Height;
                }
                graphics.ToggleFullScreen();
                metaballTarget.Dispose();
                metaballTarget = new RenderTarget2D(GraphicsDevice, GraphicsDevice.Viewport.Width, GraphicsDevice.Viewport.Height);
                var viewport = GraphicsDevice.Viewport;
                effect.Projection = Matrix.CreateTranslation(-0.5f, -0.5f, 0) * Matrix.CreateOrthographicOffCenter(0, viewport.Width, viewport.Height, 0, 0, 1);
            }

            // pressing the up or down arrows changes the alpha threshold
            if (state.IsKeyDown(Keys.Up))
                effect.ReferenceAlpha = (int)MathHelper.Clamp(effect.ReferenceAlpha + 1, 0, 256);
            if (state.IsKeyDown(Keys.Down))
                effect.ReferenceAlpha = (int)MathHelper.Clamp(effect.ReferenceAlpha - 1, 0, 256);

            // the plus and minus keys add and remove metaballs
            if (state.IsKeyDown(Keys.OemPlus))
            {
                int index = currentColor;
                if (currentColor == metaballTextures.Length)
                    index = rand.Next(metaballTextures.Length);

                balls.Add(new Metaball()
                {
                    Glow = glowColors[index],
                    Texture = metaballTextures[index],
                    Position = new Vector2(rand.Next(GraphicsDevice.Viewport.Width), rand.Next(GraphicsDevice.Viewport.Height)),
                    Velocity = rand.NextVector2(1)
                });
            }
            if (state.IsKeyDown(Keys.OemMinus) && balls.Count > 0)
                balls.RemoveAt(rand.Next(balls.Count));

            if (WasPressed(Keys.F10))
                showOverlay = !showOverlay;
        }
        
        protected override void Draw(GameTime gameTime)
        {
            // first we render the metaballs to a render target using additive blending
            GraphicsDevice.SetRenderTarget(metaballTarget);
            GraphicsDevice.Clear(Color.Transparent);

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.Additive);
            foreach (var ball in balls)
                spriteBatch.Draw(ball.Texture, ball.Position, null, Color.White, 0f, Vector2.Zero, MetaballScale, SpriteEffects.None, 0f);
            spriteBatch.End();

            // now we render everything to the backbuffer
            GraphicsDevice.SetRenderTarget(null);
            GraphicsDevice.Clear(Color.Black);

            // draw a faint glow behind the metaballs. We accomplish this by rendering the 
            // metaball texture without threshholding it. This is purely aesthetic.
            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.Additive);
            foreach (var ball in balls)
            {
                Color tint = new Color(ball.Glow.ToVector3() * 0.34f);
                spriteBatch.Draw(ball.Texture, ball.Position, null, tint, 0f, Vector2.Zero, MetaballScale, SpriteEffects.None, 0f);
            }
            spriteBatch.End();

            // now draw the metaball's render target to the screen using the AlphaTestEffect to threshhold it.
            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.Opaque, null, null, null, effect);
            spriteBatch.Draw(metaballTarget, Vector2.Zero, Color.White);
            spriteBatch.End();

#if WINDOWS_PHONE
            // draw Windows Phone instructions
            spriteBatch.Begin();
            spriteBatch.Draw(instructions, Vector2.Zero, Color.White);
            spriteBatch.End();
#else
            // draw PC instructions
            if (showOverlay)
            {
                // darken the metaballs where the menu should be drawn
                var viewport = GraphicsDevice.Viewport;
                spriteBatch.Begin();
                spriteBatch.Draw(overlay, new Vector2(100), Color.Black);
                spriteBatch.DrawString(font, "Main Menu", new Vector2(400 - font.MeasureString("Main Menu").X / 2, 140), Color.Blue);
                spriteBatch.DrawString(font, "Up and down arrows adjust the alpha threshhold.", new Vector2(140, 180), Color.Blue);
                spriteBatch.DrawString(font, "F10 toggles this overlay.", new Vector2(140, 210), Color.Blue);
                spriteBatch.DrawString(font, "C toggles colours.", new Vector2(140, 240), Color.Blue);
                spriteBatch.End();
            }
#endif

            base.Draw(gameTime);
        }


        // helper functions

        // returns whether a key was just pressed down
        bool WasPressed(Keys key)
        {
            return lastState.IsKeyUp(key) && state.IsKeyDown(key);
        }

        // returns true if either shift key is held down.
        bool Shift { get { return state.IsKeyDown(Keys.LeftShift) || state.IsKeyDown(Keys.RightShift); } }
    }
}
